from biocframe import BiocFrame
import dolomite_base as dl
import numpy as np
from tempfile import mkdtemp


def test_csv_data_frame_list():
    df = BiocFrame({
        "akari": [ 1, 2, 3, 4, 5 ],
        "aika": [ "sydney", "melbourne", "", "perth", "adelaide" ],
        "alice": [ True, False, False, True, True ],
        "ai": [ 2.3, 1.2, 5.2, 3.1, -1.2 ],
        "alicia": [ 5, 2, 1, 3.2, -2 ], # mixed integers and numbers
    })

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["columns"][0] == { "type": "integer", "name": "akari" }
    assert meta["data_frame"]["columns"][1] == { "type": "string", "name": "aika" }
    assert meta["data_frame"]["columns"][2] == { "type": "boolean", "name": "alice" }
    assert meta["data_frame"]["columns"][3] == { "type": "number", "name": "ai" }
    assert meta["data_frame"]["columns"][4] == { "type": "number", "name": "alicia" }
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)

    assert list(roundtrip.column("akari")) == df.column("akari")
    assert roundtrip.column("akari").dtype.type == np.int32

    assert roundtrip.column("aika") == df.column("aika")

    assert list(roundtrip.column("alice")) == df.column("alice")
    assert roundtrip.column("alice").dtype.type == np.bool_

    assert list(roundtrip.column("ai")) == df.column("ai")
    assert roundtrip.column("ai").dtype.type == np.float64

    assert list(roundtrip.column("alicia")) == df.column("alicia")
    assert roundtrip.column("alicia").dtype.type == np.float64


def test_csv_data_frame_row_names():
    df = BiocFrame({
        "akari": [ 1, 2, 3, 4, 5 ],
        "aika": [ "sydney", "melbourne", "", "perth", "adelaide" ],
        "alice": [ True, False, False, True, True ],
        "ai": [ 2.3, 1.2, 5.2, 3.1, -1.2 ],
    }, row_names = [ "kaori", "chihaya", "fuyuki", "azusa", "iori" ])

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["row_names"]
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.row_names == roundtrip.row_names


def test_csv_data_frame_wild_strings():
    df = BiocFrame({
        "lyrics": [ "Ochite\niku\tsunadokei\nbakari\tmiteru\tyo", "foo\"asdasd", "multie\"foo\"asdas" ],
    }, row_names = [ "nagisa\"", "\"fuko\"", "okazaki\nasdasd" ])

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["row_names"]
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.row_names == roundtrip.row_names
    assert df.column("lyrics") == roundtrip.column("lyrics")


def test_csv_data_frame_none():
    df = BiocFrame({
        "akari": [ 1, 2, None, 4, 5 ],
        "aika": [ "sydney", None, "", "perth", "adelaide" ],
        "alice": [ True, None, False, None, True ],
        "ai": [ 2.3, None, 5.2, None, -1.2 ],
    })

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["columns"][0] == { "type": "integer", "name": "akari" }
    assert meta["data_frame"]["columns"][1] == { "type": "string", "name": "aika" }
    assert meta["data_frame"]["columns"][2] == { "type": "boolean", "name": "alice" }
    assert meta["data_frame"]["columns"][3] == { "type": "number", "name": "ai" }
    dl.write_metadata(meta, dir)

    def compare_masked_to_list(l, m):
        assert len(l) == len(m)
        for i, x in enumerate(l):
            if x is not None:
                assert x == m[i]
            else:
                assert np.ma.is_masked(m[i])

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)

    compare_masked_to_list(df.column("akari"), roundtrip.column("akari"))
    assert roundtrip.column("akari").dtype.type == np.int32

    assert roundtrip.column("aika") == df.column("aika")

    compare_masked_to_list(df.column("alice"), roundtrip.column("alice"))
    assert roundtrip.column("alice").dtype.type == np.bool_

    compare_masked_to_list(df.column("ai"), roundtrip.column("ai"))
    assert roundtrip.column("ai").dtype.type == np.float64


def test_csv_data_frame_numpy():
    df = BiocFrame({
        "alicia": np.array([ 1, 2, 3, 4, 5 ]),
        "akira": np.array([ True, True, False, False, True ]),
        "athena": np.array([ 2.3, 2.3, 5.2, 32, -1.2 ]),
    })

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["columns"][0] == { "type": "integer", "name": "alicia" }
    assert meta["data_frame"]["columns"][1] == { "type": "boolean", "name": "akira" }
    assert meta["data_frame"]["columns"][2] == { "type": "number", "name": "athena" }
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert (roundtrip.column("alicia") == df.column("alicia")).all()
    assert roundtrip.column("alicia").dtype == np.int32
    assert (roundtrip.column("akira") == df.column("akira")).all()
    assert (roundtrip.column("athena") == df.column("athena")).all()


def test_csv_data_frame_masked():
    df = BiocFrame({
        "alicia": np.ma.array(np.array([ 1, 2, 3, 4, 5 ]), mask=[0, 1, 0, 1, 0]),
        "akira": np.ma.array(np.array([ True, True, False, False, True ]), mask=[1, 1, 0, 0, 0]),
        "athena": np.ma.array(np.array([ 2.3, 2.3, 5.2, 32, -1.2 ]), mask=[0, 0, 0, 1, 1]),
    })

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["columns"][0] == { "type": "integer", "name": "alicia" }
    assert meta["data_frame"]["columns"][1] == { "type": "boolean", "name": "akira" }
    assert meta["data_frame"]["columns"][2] == { "type": "number", "name": "athena" }
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert (roundtrip.column("alicia") == df.column("alicia")).all()
    assert (roundtrip.column("akira") == df.column("akira")).all()
    assert (roundtrip.column("athena") == df.column("athena")).all()


def test_csv_data_frame_empty():
    dir = mkdtemp()

    # Empty except for row names.
    df = BiocFrame({}, row_names=["chihaya", "mami", "ami", "miki", "haruka"])
    meta = dl.stage_object(df, dir, "foo")
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.row_names == roundtrip.row_names

    # Fully empty
    df = BiocFrame(number_of_rows=10)
    meta = dl.stage_object(df, dir, "empty")
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "empty/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.shape == roundtrip.shape
    assert roundtrip.row_names is None


def test_csv_data_frame_nested():
    df = BiocFrame({
        "odd": [ 1, 3, 5, 7, 9 ],
        "liella": BiocFrame({ 
            "first": [ "kanon", "keke", "chisato", "sumire", "ren" ],
            "last": [ "shibuya", "tang", "arashi", "heanna", "hazuki" ],
            "best": [ False, False, False, True, False ],
        }),
        "even": [ 0, 2, 4, 6, 8 ],
        "bsb": BiocFrame({ 
            "first": [ "nick", "kevin", "brian", "AJ", "howie" ],
            "last": [ "carter", "richardson", "litrell", "maclean", "dorough" ]
        }),
    })

    dir = mkdtemp()
    meta = dl.stage_object(df, dir, "foo")
    assert meta["data_frame"]["columns"][1]["type"] == "other"
    assert meta["data_frame"]["columns"][3]["type"] == "other"
    dl.write_metadata(meta, dir)

    meta2 = dl.acquire_metadata(dir, "foo/simple.csv.gz")
    roundtrip = dl.load_object(meta2, dir)
    assert isinstance(roundtrip, BiocFrame)

    liella_df = roundtrip.column("liella")
    assert isinstance(liella_df, BiocFrame)
    assert liella_df.column("first") == df.column("liella").column("first")
    assert liella_df.column("last") == df.column("liella").column("last")
    assert (liella_df.column("best") == np.array([False, False, False, True, False ])).all()

    bsb_df = roundtrip.column("bsb")
    assert isinstance(bsb_df, BiocFrame)
    assert bsb_df.column("first") == df.column("bsb").column("first")
    assert bsb_df.column("last") == df.column("bsb").column("last")



from typing import Any
import os
from biocframe import BiocFrame

from .stage_object import stage_object
from ._stage_csv_data_frame import _stage_csv_data_frame
from .write_metadata import write_metadata


@stage_object.register
def stage_data_frame(x: BiocFrame, dir: str, path: str, is_child: bool = False, **kwargs) -> dict[str, Any]:
    """Method for saving :py:class:`~biocframe.BiocFrame.BiocFrame`
    objects to the corresponding file representations, see
    :py:meth:`~dolomite_base.stage_object.stage_object` for details.

    Args:
        x: Object to be staged.

        dir: Staging directory.

        path: Relative path inside ``dir`` to save the object.

        is_child: Is ``x`` a child of another object?

        kwargs: Further arguments, ignored.

    Returns:
        Metadata that can be edited by calling methods and then saved with 
        :py:meth:`~dolomite_base.write_metadata.write_metadata`.
    """
    os.mkdir(os.path.join(dir, path))
    meta, other = _stage_csv_data_frame(x, dir, path, is_child=is_child)

    for i in other:
        more_meta = stage_object(x.column(i), dir, path + "/child-" + str(i + 1), is_child = True)
        resource_stub = write_metadata(more_meta, dir=dir)
        meta["data_frame"]["columns"][i]["resource"] = resource_stub

    return meta



import ctypes as ct
from typing import Any
from biocframe import BiocFrame
import numpy as np
import os

from . import _cpphelpers as lib
from .acquire_file import acquire_file
from .acquire_metadata import acquire_metadata
from .load_object import load_object


class _LoadedCsvHolder:
    def __init__(self, ptr):
        self.ptr = ptr
        self._numf = None
        self._numr = None

    def __del__(self):
        lib.free_csv(self.ptr)

    def num_fields(self):
        if self._numf is None:
            self._numf = lib.get_csv_num_fields(self.ptr)
        return self._numf

    def num_records(self):
        if self._numr is None:
            self._numr = lib.get_csv_num_records(self.ptr)
        return self._numr

    def column(self, i: int, pop: bool = False):
        col_type = ct.c_int32(0)
        col_size = ct.c_int32(0)
        col_loaded = ct.c_int32(0)
        lib.get_csv_column_stats(self.ptr, i, ct.byref(col_type), ct.byref(col_size), ct.byref(col_loaded))

        if col_loaded.value == 0:
            return None
        N = col_size.value
        if N != self.num_records():
            raise ValueError("column size exceeds the number of records in the CSV")

        if col_type.value == 0:
            strlengths = np.ndarray(N, dtype=np.int32)
            mask = np.zeros(N, dtype=np.uint8)
            lib.get_csv_string_stats(self.ptr, i, strlengths, mask)

            total_len = int(strlengths.sum())
            concatenated = ct.create_string_buffer(total_len)
            lib.fetch_csv_strings(self.ptr, i, concatenated, pop)

            sofar = 0
            collected = []
            buffer = concatenated.raw
            for i, x in enumerate(strlengths):
                endpoint = sofar + x 
                collected.append(buffer[sofar:endpoint].decode("ASCII"))
                sofar = endpoint

            for i, x in enumerate(mask):
                if x:
                    collected[i] = None
            return collected

        elif col_type.value == 1:
            values = np.ndarray(N, dtype=np.float64)
            mask = np.zeros(N, dtype=np.uint8)
            masked = lib.fetch_csv_numbers(self.ptr, i, values, mask, pop)
            if masked:
                return np.ma.array(values, mask=mask)
            else:
                return values

        elif col_type.value == 3:
            values = np.ndarray(N, dtype=np.uint8)
            masked = lib.fetch_csv_booleans(self.ptr, i, values, pop)
            if masked:
                mask = values == 2
                return np.ma.array(values.astype(dtype=np.bool_), mask=mask)
            else:
                return values.astype(dtype=np.bool_)

        elif col_type.value == -1:
            return None

        else:
            return NotImplementedError("not-yet-supported type for column " + str(i) + " of the CSV (" + str(col_type.value) + ")")


def load_csv_data_frame(meta: dict[str, Any], project: Any, **kwargs) -> BiocFrame:
    """Load a data frame from a (possibly Gzip-compressed) CSV file in the
    **comservatory** format. In general, this function should not be called
    directly but instead via :py:meth:`~dolomite_base.load_object.load_object`.

    Args:
        meta: Metadata for this CSV data frame.

        project: Value specifying the project of interest. This is most
            typically a string containing a file path to a staging directory
            but may also be an application-specific object that works with
            :py:meth:`~dolomite_base.acquire_file.acquire_file`.

        kwargs: Further arguments, passed to nested objects.

    Returns:
        A data frame.
    """
    full_path = acquire_file(project, meta["path"])
    handle = _LoadedCsvHolder(lib.load_csv(full_path.encode("UTF8")))

    has_row_names = "row_names" in meta["data_frame"] and meta["data_frame"]["row_names"]
    columns = meta["data_frame"]["columns"]

    expected_cols = len(columns) + has_row_names 
    observed_cols = handle.num_fields()
    if expected_cols != observed_cols:
        raise ValueError("difference between the observed and expected number of CSV columns (" + str(observed_cols) + " to " + str(expected_cols) + ")")

    expected_rows = meta["data_frame"]["dimensions"][0]
    observed_rows = handle.num_records()
    if expected_rows != observed_rows:
        raise ValueError("difference between the observed and expected number of CSV rows (" + str(observed_rows) + " to " + str(expected_rows) + ")")

    contents = []
    row_names = None
    for f in range(observed_cols):
        # Always popping columns from the C++ representation as soon as we're
        # done, so as to free up memory. Not really sure whether this has much
        # of an effect as the C++/Python heaps aren't the same.
        current = handle.column(f, pop=True)
        if f == 0 and has_row_names:
            row_names = current
        else:
            contents.append(current)

    output = BiocFrame({}, number_of_rows=expected_rows, row_names=row_names)
    for i, c in enumerate(contents):
        curval = columns[i]
        if curval["type"] == "other":
            child_meta = acquire_metadata(project, curval["resource"]["path"])
            c = load_object(child_meta, project, **kwargs)
        elif curval["type"] == "integer":
            c = c.astype(np.int32)
        output[curval["name"]] = c
   
    return output



from typing import Any
from importlib import import_module
import os

from ._schemas import _hunt_for_schemas


def custom_load_object_helper(meta: dict, project: Any, locations: list, memory: dict[str, Any], **kwargs) -> Any:
    """Helper function to create application-specific variants of
    :py:meth:`~dolomite_base.load_object.load_object`. This allows
    applications to divert to different schema locations and/or
    loading functions for one or multiple object classes.

    Args:
        meta: Metadata for this object.

        project: 
            Value specifying the project of interest. This is most
            typically a string containing a file path to a staging directory
            but may also be an application-specific object that works with
            :py:meth:`~dolomite_base.acquire_file.acquire_file`.

        locations: 
            Application-specific list of names of packages to be searched for
            object schemas. It is expected that the schema for each object
            specifies the function required to create that object from its file
            representations.

        memory:
            Application-specific cache of restoration functions for each object
            schema. This avoids redundant look-ups if the same ``memory`` is
            recycled across calls.

        kwargs: Further arguments, passed to individual methods.

    Returns:
        Some kind of object.
    """
    schema = meta["$schema"]

    if schema not in memory:
        schema_details = _hunt_for_schemas(locations, schema)
        command = None
        if "_attributes" in schema_details:
            attr_meta = schema_details["_attributes"]
            if "restore" in attr_meta:
                res_meta = attr_meta["restore"]
                if "python" in res_meta:
                    command = res_meta["python"]

        if command is None:
            # TODO: replace these in the schemas themselves, once everything has settled down.
            if schema == "csv_data_frame/v1.json":
                command = "dolomite_base.load_csv_data_frame"
            else:
                raise NotImplementedError("could not find a Python context to restore '" + schema + "'")

        first_period = command.find(".")
        mod = import_module(command[:first_period])
        memory[schema] = getattr(mod, command[first_period + 1:])

    return memory[schema](meta, project, **kwargs)


#: Default locations (as package names) in which to search for schemas.
#: Each package should have a ``schemas/`` subdirectory containing all schemas.
#: Advanced users can insert their own packages here to influence the search.
DEFAULT_SCHEMA_LOCATIONS = [ "dolomite_schemas" ]


_schema_restoration = {}


def load_object(meta: dict, project: Any, **kwargs) -> Any:
    """Load an object from a resource inside a project. This uses the schemas
    in :py:obj:`~DEFAULT_SCHEMA_LOCATIONS` to identify the restoration functions.

    Args:
        meta: Metadata for this object.

        project: Value specifying the project of interest. This is most
            typically a string containing a file path to a staging directory
            but may also be an application-specific object that can be handled
            by :py:meth:`~dolomite_base.acquire_file.acquire_file`.

        kwargs: Further arguments, passed to individual methods.

    Returns:
        Some kind of object.
    """
    return custom_load_object_helper(
        meta, 
        project, 
        locations = DEFAULT_SCHEMA_LOCATIONS, 
        memory = _schema_restoration, 
        **kwargs
    ) 



from typing import Any
from functools import singledispatch


@singledispatch
def stage_object(x: Any, dir: str, path: str, **kwargs) -> dict[str, Any]:
    """Save an object to file in a staging directory. **dolomite** extensions
    should define methods for this generic to stage different object classes.

    Arguments:
        x: Object to be staged.

        dir: Path to a staging directory.

        path: Relative path inside ``dir`` where ``x`` is to be saved.
            This will be used to create a subdirectory inside ``dir``.

        kwargs: Further arguments to be passed to individual methods.

    Returns:
        Metadata for ``x``, which can be edited by calling methods and then
        saved with :py:meth:`~dolomite_base.write_metadata.write_metadata`.
        This should contain at least a ``$schema`` property, specifying the
        schema to use for validation of the metadata; and a ``path`` property,
        specifying the relative path to the file representation of ``x``.
    """
    raise NotImplementedError("'stage_object' is not implemented for " + str(type(x)))



import dolomite_base as dl
import tempfile
import json
import os


def _mockup(tmpdir, basename): 
    with open(os.path.join(tmpdir, basename), "w") as handle:
        handle.write("your mom")

    return {
        "$schema": "generic_file/v1.json",
        "path": basename,
        "generic_file": {
            "format": "text/plain"
        }
    }


def test_write_metadata_basic():
    tmpdir = tempfile.mkdtemp()
    basename = "foo.txt"
    full_meta = _mockup(tmpdir, basename)

    out = dl.write_metadata(full_meta, tmpdir)
    assert out["type"] == "local"
    assert out["path"] == basename

    with open(os.path.join(tmpdir, basename + ".json"), "r") as handle:
        roundtrip = json.load(handle)
    assert isinstance(roundtrip["md5sum"], str)
    assert roundtrip["generic_file"]["format"] == "text/plain"


def test_write_metadata_none():
    tmpdir = tempfile.mkdtemp()
    basename = "foo.txt"

    full_meta = _mockup(tmpdir, basename)
    full_meta["generic_file"]["foo"] = None

    # Nones are automatically stripped out.
    out = dl.write_metadata(full_meta, tmpdir)
    with open(os.path.join(tmpdir, basename + ".json"), "r") as handle:
        roundtrip = json.load(handle)
    assert "foo" not in roundtrip["generic_file"]


def test_write_metadata_metaonly():
    tmpdir = tempfile.mkdtemp()
    os.mkdir(os.path.join(tmpdir, "yay"))

    basename = "yay/whee.json"
    full_meta = {
        "$schema": "summarized_experiment/v1.json",
        "path": basename,
        "summarized_experiment": {
            "dimensions": [ 100, 10 ],
            "assays": [
                {
                    "name": "counts",
                    "resource": {
                        "type": "local",
                        "path": "yay/assay-1/simple.h5"
                    }
                } 
            ]
        }
    }

    # Nones are automatically stripped out.
    out = dl.write_metadata(full_meta, tmpdir)
    with open(os.path.join(tmpdir, basename), "r") as handle:
        roundtrip = json.load(handle)
    assert roundtrip == full_meta
    assert "md5sum" not in roundtrip


def test_write_metadata_sanitized():
    tmpdir = tempfile.mkdtemp()
    basename = "./foo.txt"
    full_meta = _mockup(tmpdir, basename)

    # We automatically remove the preceding './'.
    out = dl.write_metadata(full_meta, tmpdir)
    with open(os.path.join(tmpdir, basename + ".json"), "r") as handle:
        roundtrip = json.load(handle)
    assert roundtrip["path"] == "foo.txt"


def test_write_metadata_tuple():
    tmpdir = tempfile.mkdtemp()
    os.mkdir(os.path.join(tmpdir, "stuff"))

    basename = "./stuff/foo.txt"
    full_meta = _mockup(tmpdir, basename)
    full_meta["$schema"] = (full_meta["$schema"], "dolomite_schemas")

    out = dl.write_metadata(full_meta, tmpdir)
    with open(os.path.join(tmpdir, basename + ".json"), "r") as handle:
        roundtrip = json.load(handle)
    assert roundtrip["$schema"] == "generic_file/v1.json"
    assert roundtrip["path"] == "stuff/foo.txt"



from typing import Any, Tuple
from biocframe import BiocFrame
import numpy as np
import os
import gzip
from . import _cpphelpers as lib


def _list_element_to_string(s):
    if s is None:
        return "NA"
    return str(s)


def _numpy_element_to_string(s):
    if np.ma.is_masked(s):
        return "NA"
    return str(s)


def _quotify_string(s):
    if '"' in s:
        s = s.replace('"', '""')
    return '"' + s + '"'


def _quotify_string_or_none(s):
    if s is None:
        return "NA"
    return _quotify_string(s)


def _process_columns(x: BiocFrame) -> Tuple:
    columns = []
    otherable = []
    operations = []

    # TODO: handle date-times, pandas' Categorical factors.
    for i, col in enumerate(x.column_names):
        current = x.column(col)
        redirect = False
        final_type = bool

        if isinstance(current, list):
            all_types = set([type(y) for y in current])
            if type(None) in all_types:
                all_types.remove(type(None))

            if len(all_types) == 1:
                final_type = list(all_types)[0]
            elif len(all_types) == 2 and int in all_types and float in all_types:
                final_type = float
            else:
                redirect = True

            if redirect:
                columns.append({ "type": "other", "name": col })
                operations.append(lambda x : 0)
                otherable.append(i)
            elif final_type == int:
                columns.append({ "type": "integer", "name": col })
                operations.append(_list_element_to_string)
            elif final_type == float:
                columns.append({ "type": "number", "name": col })
                operations.append(_list_element_to_string)
            elif final_type == str:
                columns.append({ "type": "string", "name": col })
                operations.append(_quotify_string_or_none)
            elif final_type == bool:
                columns.append({ "type": "boolean", "name": col })
                operations.append(_list_element_to_string)
            else:
                raise NotImplementedError("saving a list of " + str(list(all_types)[0]) + " is not supported yet")

        elif isinstance(current, np.ndarray):
            dt = current.dtype.type
            if issubclass(dt, np.integer):
                columns.append({ "type": "integer", "name": col })
            elif issubclass(dt, np.floating):
                columns.append({ "type": "number", "name": col })
            elif issubclass(dt, np.bool_):
                columns.append({ "type": "boolean", "name": col })
            else:
                raise NotImplementedError("saving a NumPy array of " + str(current.dtype) + " is not supported yet")

            if np.ma.is_masked(current):
                operations.append(_numpy_element_to_string)
            else:
                operations.append(str)

        else:
            columns.append({ "type": "other", "name": col })
            operations.append(lambda x : "0")
            otherable.append(i)

    return columns, otherable, operations


def _stage_csv_data_frame(x: BiocFrame, dir: str, path: str, is_child: bool) -> Tuple:
    columns, otherable, operations = _process_columns(x)
    nr = x.shape[0]

    extracted_row_names = x.row_names
    print(x)
    has_row_names = extracted_row_names is not None
    extracted_columns = []
    for y in x.column_names:
        extracted_columns.append(x.column(y))

    # Manual serialization into a Gzip-compressed CSV, because 
    # pandas doesn't quite give me what I want... oh well.
    basename = "simple.csv.gz"
    full = os.path.join(dir, path, basename)

    with gzip.open(full, "wb") as handle:
        header_line = ""
        if has_row_names:
            header_line += _quotify_string("row_names")
        for c in x.column_names:
            if header_line:
                header_line += ","
            header_line += _quotify_string(c)
        header_line += "\n"
        handle.write(header_line.encode("ASCII"))

        for r in range(nr):
            if has_row_names:
                line = _quotify_string(extracted_row_names[r])
            else:
                line = ""
            for i, trans in enumerate(operations):
                if line:
                    line += ","
                line += trans(extracted_columns[i][r])
            line += "\n"
            handle.write(line.encode("ASCII"))

    lib.validate_csv(full.encode("UTF8"))
    metadata = {
        "$schema": "csv_data_frame/v1.json",
        "path": path + "/" + basename,
        "is_child": is_child,
        "data_frame": {
            "columns": columns,
            "row_names": x.row_names is not None,
            "dimensions": list(x.shape),
        },
        "csv_data_frame": {
            "compression": "gzip",
        }
    }

    return metadata, otherable



from typing import Any
from functools import singledispatch
import os
import json


@singledispatch
def acquire_metadata(project: Any, path: str) -> dict[str, Any]:
    """Acquire metadata for a resource inside a project. Applications should
    define methods for this generic to acquire metadata from different places.

    Args:
        project: 
            Any value that specifies the project of interest.  By default, this
            should be a path to a staging directory, but other applications may
            provide their own classes, e.g., to identify projects on remotes. 

        path:
            Relative path to the resource of interest inside the project.

    Returns:
        Dictionary representing the JSON metadata.
    """
    raise NotImplementedError("'acquire_metadata' for " + str(type(project)) + " has not been implemented")


@acquire_metadata.register
def acquire_metadata_from_dir(project: str, path: str) -> str:
    """Acquire metadata for a resource inside a local staging directory.
    This is a method for the :py:meth:`~acquire_metadata` generic.

    Args:
        project: 
            Path to a staging directory.

        path:
            Relative path to the resource of interest inside the project.

    Returns:
        Dictionary representing the JSON metadata.
    """
    if not path.endswith(".json"):
        path += ".json"
    fpath = os.path.join(project, path)
    with open(fpath, "r") as handle:
        return json.load(handle)



# DO NOT MODIFY: this is automatically generated by the cpptypes

import os
import ctypes as ct
import numpy as np

def _catch_errors(f):
    def wrapper(*args):
        errcode = ct.c_int32(0)
        errmsg = ct.c_char_p(0)
        output = f(*args, ct.byref(errcode), ct.byref(errmsg))
        if errcode.value != 0:
            msg = errmsg.value.decode('ascii')
            lib.free_error_message(errmsg)
            raise RuntimeError(msg)
        return output
    return wrapper

# TODO: surely there's a better way than whatever this is.
dirname = os.path.dirname(os.path.abspath(__file__))
contents = os.listdir(dirname)
lib = None
for x in contents:
    if x.startswith('_core') and not x.endswith("py"):
        lib = ct.CDLL(os.path.join(dirname, x))
        break

if lib is None:
    raise ImportError("failed to find the _core.* module")

lib.free_error_message.argtypes = [ ct.POINTER(ct.c_char_p) ]

def _np2ct(x, expected, contiguous=True):
    if not isinstance(x, np.ndarray):
        raise ValueError('expected a NumPy array')
    if x.dtype != expected:
        raise ValueError('expected a NumPy array of type ' + str(expected) + ', got ' + str(x.dtype))
    if contiguous:
        if not x.flags.c_contiguous and not x.flags.f_contiguous:
            raise ValueError('only contiguous NumPy arrays are supported')
    return x.ctypes.data

lib.py_fetch_csv_booleans.restype = ct.c_uint8
lib.py_fetch_csv_booleans.argtypes = [
    ct.c_void_p,
    ct.c_int32,
    ct.c_void_p,
    ct.c_uint8,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_fetch_csv_numbers.restype = ct.c_uint8
lib.py_fetch_csv_numbers.argtypes = [
    ct.c_void_p,
    ct.c_int32,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_uint8,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_fetch_csv_strings.restype = None
lib.py_fetch_csv_strings.argtypes = [
    ct.c_void_p,
    ct.c_int32,
    ct.c_char_p,
    ct.c_uint8,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_free_csv.restype = None
lib.py_free_csv.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_get_csv_column_stats.restype = None
lib.py_get_csv_column_stats.argtypes = [
    ct.c_void_p,
    ct.c_int32,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_get_csv_num_fields.restype = ct.c_int32
lib.py_get_csv_num_fields.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_get_csv_num_records.restype = ct.c_int32
lib.py_get_csv_num_records.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_get_csv_string_stats.restype = ct.c_uint8
lib.py_get_csv_string_stats.argtypes = [
    ct.c_void_p,
    ct.c_int32,
    ct.c_void_p,
    ct.c_void_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_load_csv.restype = ct.c_void_p
lib.py_load_csv.argtypes = [
    ct.c_char_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

lib.py_validate_csv.restype = None
lib.py_validate_csv.argtypes = [
    ct.c_char_p,
    ct.POINTER(ct.c_int32),
    ct.POINTER(ct.c_char_p)
]

def fetch_csv_booleans(ptr, column, contents, pop):
    return _catch_errors(lib.py_fetch_csv_booleans)(ptr, column, _np2ct(contents, np.uint8), pop)

def fetch_csv_numbers(ptr, column, contents, mask, pop):
    return _catch_errors(lib.py_fetch_csv_numbers)(ptr, column, _np2ct(contents, np.float64), _np2ct(mask, np.uint8), pop)

def fetch_csv_strings(ptr, column, contents, pop):
    return _catch_errors(lib.py_fetch_csv_strings)(ptr, column, contents, pop)

def free_csv(ptr):
    return _catch_errors(lib.py_free_csv)(ptr)

def get_csv_column_stats(ptr, column, type, size, loaded):
    return _catch_errors(lib.py_get_csv_column_stats)(ptr, column, type, size, loaded)

def get_csv_num_fields(ptr):
    return _catch_errors(lib.py_get_csv_num_fields)(ptr)

def get_csv_num_records(ptr):
    return _catch_errors(lib.py_get_csv_num_records)(ptr)

def get_csv_string_stats(ptr, column, lengths, mask):
    return _catch_errors(lib.py_get_csv_string_stats)(ptr, column, _np2ct(lengths, np.int32), _np2ct(mask, np.uint8))

def load_csv(path):
    return _catch_errors(lib.py_load_csv)(path)

def validate_csv(path):
    return _catch_errors(lib.py_validate_csv)(path)


from typing import Any
from functools import singledispatch
import os


@singledispatch
def acquire_file(project: Any, path : str) -> str:
    """Acquire the file path for a resource. Applications should define methods
    for this generic to acquire metadata from other places.

    Arguments:
        project: 
            Any value that specifies the project of interest.  By default, this
            should be a path to a staging directory, but other applications may
            provide their own classes, e.g., identify projects on remotes. 

        path:
            Relative path to the resource of interest inside the project.

    Returns:
        File path to the resource, possibly after downloading and caching.
    """
    raise NotImplementedError("'acquire_file' for " + str(type(project)) + " has not been implemented")


@acquire_file.register
def acquire_file_from_dir(project: str, path: str) -> str:
    """Acquire the file path for a resource in a local staging directory.
    This is a method for the :py:meth:`~acquire_file` generic.

    Arguments:
        project: 
            Path to a staging directory.

        path:
            Relative path to the resource of interest inside the project.

    Returns:
        File path to the resource, possibly after downloading and caching.
    """
    return os.path.join(project, path)



"""
    Setup file for dolomite_base.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 4.5.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
from setuptools import setup, Extension
import assorthead
from glob import glob


if __name__ == "__main__":
    import os
    try:
        setup(
            use_scm_version={"version_scheme": "no-guess-dev"},
            ext_modules=[
                Extension(
                    "dolomite_base._core",
                    sorted(glob("src/dolomite_base/lib/*.cpp")),
                    include_dirs=[
                        assorthead.includes(),
                        "src/dolomite_base/include"
                    ],
                    language="c++",
                    extra_compile_args=[
                        "-std=c++17",
                    ],
                    extra_link_args=[
                        "-lz"
                    ],
                )
            ]
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



import json
from importlib import import_module 
import os

_schema_cache = {}

def _fetch_schema(pkg, schema):
    if pkg not in _schema_cache:
        _schema_cache[pkg] = {}

    current_cache = _schema_cache[pkg]
    if schema not in current_cache:
        schema_pkg = import_module(pkg)
        schema_path = os.path.join(os.path.dirname(schema_pkg.__file__), "schemas", schema)
        if os.path.exists(schema_path):
            with open(schema_path, "r") as handle:
                current_cache[schema] = json.load(handle)
        else:
            current_cache[schema] = None

    return current_cache[schema]


def _hunt_for_schemas(locations, schema, fallback = None):
    schema_details = None
    for pkg in locations:
        schema_details = _fetch_schema(pkg, schema)
        if schema_details is not None:
            break

    if schema_details is None:
        if fallback is None:
            raise FileNotFoundError("could not find schema '" + schema + "' in any package location")
        schema_details = fallback(schema)

    return schema_details



from typing import Any
from hashlib import md5
import os
import json
from jsonschema import validate
from ._schemas import _fetch_schema
from copy import copy


def _strip_none_values(x):
    if isinstance(x, dict):
        kill = []
        for k, v in x.items():
            if isinstance(v, dict) or isinstance(v, list):
                _strip_none_values(v)
            elif v is None:
                kill.append(k)
        if len(kill):
            for k in kill:
                del x[k]
    elif isinstance(x, list):
        for v in x:
            if isinstance(v, dict) or isinstance(v, list):
                _strip_none_values(v)


def write_metadata(meta: dict[str, Any], dir: str, ignore_none: bool = True) -> dict[str, str]:
    """Write metadata to disk after validating it against its schema. 

    ``meta`` should contain at least the ``$schema`` and ``path`` keys. 
    ``$schema`` is used to identify the schema for validation; it should be
    a string or a 2-tuple containing the schema name and the name of the 
    package in which the schema file can be found.

    If the metadata describes a resource with a file representation, that file
    should already be present inside ``dir`` at the specified ``path``.  Also,
    if no ``md5sum`` key is present,  the MD5 checksum of the file at ``path``
    is computed and added to the metadata before saving.

    Any preceding ``./`` in the ``path`` is automatically removed.

    Args:
        meta: 
            Metadata for an object, typically generated by a
            :py:meth:`~dolomite_base.stage_object.stage_object` method.

        dir:
            Path to the staging directory.

        ignore_none:
            Whether keys with None values should be automatically removed from
            ``meta``. This is useful for sanitizing messy user-supplied inputs.

    Returns:
        Dictionary that can be added as the ``resource`` property of parent
        metadata, if the current object is a child of another object. The
        metadata itself is written to a JSON file at a path derived from
        (or equal to) ``meta["path"]``.
    """
    meta = copy(meta) # make an internal copy.
    if ignore_none:
        _strip_none_values(meta)

    schema = meta["$schema"]
    if isinstance(schema, tuple):
        meta["$schema"] = schema[0]
        pkg = schema[1]
        schema = schema[0]
    else:
        pkg = "dolomite_schemas"
    schema_details = _fetch_schema(pkg, schema)

    if meta["path"].startswith("./"): # removing for convenience
        meta["path"] = meta["path"][2:]

    meta_only = False
    if "_attributes" in schema_details:
        attributes = schema_details["_attributes"]
        if "metadata_only" in attributes:
            meta_only = attributes["metadata_only"]

    jpath = meta["path"]
    if not meta_only:
        jpath += ".json"
        if "md5sum" not in meta and not schema.startswith("redirection/"):
            with open(os.path.join(dir, meta["path"]), "rb") as handle:
                hasher = md5()
                while True:
                    chunk = handle.read(65536)
                    if not chunk:
                        break
                    hasher.update(chunk)
                meta["md5sum"] = hasher.hexdigest()

    validate(meta, schema_details)
    with open(os.path.join(dir, jpath), "w") as handle:
        json.dump(meta, handle)

    return {
        "type": "local",
        "path": meta["path"]
    }
